/**********************************************************************************************
* This file is part of the Itty Bitty Arduino library.                                        *
* Copyright  2016 Thomas J. Biuso III  ALL RIGHTS RESERVED...WHATEVER THAT MEANS.            *
* RELEASED UNDER THE GPL v3.0 LICENSE; SEE <LICENSE> FILE WITHIN DISTRIBUTION ROOT FOR TERMS. *
***********************************************************************************************/

#if !defined(ARDUINO) || defined(NO_ITTYBITTY_LCD) || defined(NO_ITTYBITTY_LCD_I2C)
	#define NO_ITTYBITTY_CHARS
#endif

#ifndef NO_ITTYBITTY_CHARS


#include "IttyBitty_LCD_I2C.h"


namespace IttyBitty
{
#pragma region CUSTOM LCD CHARACTER FLASH DATA

	// Miscellaneous symbols
	PROGMEM CBYTE LCD_CHAR_CHECKMARK[8]				= { 0x0, 0x0, 0x1, 0x3, 0x16, 0x1C, 0x8, 0x0 };
	PROGMEM CBYTE LCD_CHAR_BOX_EMPTY[8]				= { 0x0, 0x1F, 0x11, 0x11, 0x11, 0x11, 0x11, 0x1F };
	PROGMEM CBYTE LCD_CHAR_BOX_CHECKMARK[8]			= { 0x0, 0x1F, 0x11, 0x13, 0x1F, 0x15, 0x11, 0x1F };
	PROGMEM CBYTE LCD_CHAR_WARNING[8]				= { 0x0, 0xE, 0x1B, 0x1B, 0x1F, 0x1B, 0xE, 0x0 };
	PROGMEM CBYTE LCD_CHAR_NOT_ALLOWED[8]			= { 0xE, 0x1F, 0x19, 0x1D, 0x17, 0x13, 0x1F, 0xE };
	PROGMEM CBYTE LCD_CHAR_LIGHTNING[8]				= { 0x3, 0x6, 0xC, 0xF, 0x3, 0x6, 0xC, 0x8 };
	PROGMEM CBYTE LCD_CHAR_NOTES[8]					= { 0x3, 0x7, 0xD, 0x9, 0x9, 0xB, 0x1B, 0x18 };
	PROGMEM CBYTE LCD_CHAR_CLOCK_A[8]				= { 0x7, 0xC, 0xA, 0x9, 0x8, 0x8, 0x7, 0x0 };
	PROGMEM CBYTE LCD_CHAR_CLOCK_B[8]				= { 0x1C, 0x2, 0x2, 0x1A, 0x2, 0x2, 0x1C, 0x0 };
	PROGMEM CBYTE LCD_CHAR_HOURGLASS_1[8]			= { 0x1F, 0x1F, 0xE, 0x4, 0xA, 0x11, 0x11, 0x1F };
	PROGMEM CBYTE LCD_CHAR_HOURGLASS_2[8]			= { 0x1F, 0x11, 0xE, 0x4, 0xA, 0x11, 0x1F, 0x1F };
	PROGMEM CBYTE LCD_CHAR_HOURGLASS_3[8]			= { 0x1F, 0x11, 0xA, 0x4, 0xA, 0x1F, 0x1F, 0x1F };
	PROGMEM CBYTE LCD_CHAR_HOURGLASS_4[8]			= { 0x1F, 0x11, 0xA, 0x4, 0xE, 0x1F, 0x1F, 0x1F };
	PROGMEM CBYTE LCD_CHAR_ANGLE[8]					= { 0x0, 0x1, 0x2, 0x4, 0xA, 0x12, 0x1F, 0x0 };
	PROGMEM CBYTE LCD_CHAR_ANGLE_2[8]				= { 0x1, 0x3, 0x6, 0xC, 0x1A, 0x12, 0x1F, 0x0 };
	PROGMEM CBYTE LCD_CHAR_CLOCKWISE[8]				= { 0x0, 0xE, 0x1B, 0x11, 0x5, 0xF, 0x4, 0x0 };
	PROGMEM CBYTE LCD_CHAR_COUNTER_CCW[8]			= { 0x0, 0xE, 0x1B, 0x11, 0x14, 0x1E, 0x4, 0x0 };


	// Shape (outline) icons
	PROGMEM CBYTE LCD_CHAR_SHAPE_CIRCLE[8]			= { 0x0, 0x0, 0xE, 0x11, 0x11, 0x11, 0xE, 0x0 };
	PROGMEM CBYTE LCD_CHAR_SHAPE_ROUND[8]			= { 0x0, 0x0, 0xE, 0x1B, 0x11, 0x1B, 0xE, 0x0 };
	PROGMEM CBYTE LCD_CHAR_SHAPE_OVAL[8]			= { 0x0, 0xE, 0x1B, 0x11, 0x11, 0x1B, 0xE, 0x0 };
	PROGMEM CBYTE LCD_CHAR_SHAPE_CUSHION[8]			= { 0x0, 0x6, 0xF, 0x19, 0x13, 0x1E, 0xC, 0x0 };
	PROGMEM CBYTE LCD_CHAR_SHAPE_SQUARE[8]			= { 0x0, 0x1F, 0x11, 0x11, 0x11, 0x11, 0x1F, 0x0 };
	PROGMEM CBYTE LCD_CHAR_SHAPE_DIAMOND[8]			= { 0x0, 0x0, 0x4, 0xA, 0x11, 0xA, 0x4, 0x0 };
	PROGMEM CBYTE LCD_CHAR_SHAPE_BAGUETTE[8]		= { 0x0, 0xE, 0xA, 0xA, 0xA, 0xA, 0xE, 0x0 };
	PROGMEM CBYTE LCD_CHAR_SHAPE_EMERALD[8]			= { 0xE, 0x1B, 0x11, 0x11, 0x11, 0x11, 0x1B, 0xE };
	PROGMEM CBYTE LCD_CHAR_SHAPE_TRILLIANT[8]		= { 0x0, 0x4, 0xE, 0x1B, 0x11, 0x1B, 0xE, 0x0 };
	PROGMEM CBYTE LCD_CHAR_SHAPE_TRIANGLE[8]		= { 0x0, 0x4, 0xE, 0xA, 0x1B, 0x11, 0x1F, 0x0 };
	PROGMEM CBYTE LCD_CHAR_SHAPE_SPECIAL[8]			= { 0x0, 0x0, 0xA, 0x15, 0x11, 0xA, 0x4, 0x0 };


	// Small arrow icons
	PROGMEM CBYTE LCD_CHAR_ARROW_UP_SMALL[8]		= { 0x0, 0x0, 0x4, 0xE, 0x4, 0x4, 0x0, 0x0 };
	PROGMEM CBYTE LCD_CHAR_ARROW_DOWN_SMALL[8]		= { 0x0, 0x0, 0x4, 0x4, 0xE, 0x4, 0x0, 0x0 };
	PROGMEM CBYTE LCD_CHAR_ARROW_LEFT_SMALL[8]		= { 0x0, 0x0, 0x0, 0x8, 0x1E, 0x8, 0x0, 0x0 };
	PROGMEM CBYTE LCD_CHAR_ARROW_RIGHT_SMALL[8]		= { 0x0, 0x0, 0x0, 0x2, 0xF, 0x2, 0x0, 0x0 };

	// Medium arrow icons
	PROGMEM CBYTE LCD_CHAR_ARROW_UP[8]				= { 0x0, 0x4, 0xE, 0x15, 0x4, 0x4, 0x4, 0x0 };
	PROGMEM CBYTE LCD_CHAR_ARROW_DOWN[8]			= { 0x0, 0x4, 0x4, 0x4, 0x15, 0xE, 0x4, 0x0 };

	// Large arrow icons
	PROGMEM CBYTE LCD_CHAR_ARROW_LEFT_LARGE[8]		= { 0x0, 0x0, 0x4, 0xC, 0x1F, 0xC, 0x4, 0x0 };
	PROGMEM CBYTE LCD_CHAR_ARROW_RIGHT_LARGE[8]		= { 0x0, 0x0, 0x4, 0x6, 0x1F, 0x6, 0x4, 0x0 };

#pragma endregion
};

#endif	// #ifndef NO_ITTYBITTY_CHARS
